/* This file was generated by plugin 'Nordic Semiconductor nRF5x v.1.2.2' (BDS version 1.1.3139.0) */

#include "ble_fas.h"
#include <string.h>
#include "nordic_common.h"
#include "ble_srv_common.h"
//#include "ble_common.h"
#include "app_util.h"
#include "app_util_bds.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "app_timer.h"



#define BLE_L2CAP_MTU_DEF               23
#define OPCODE_LENGTH 1 /**< Length of opcode inside FAS packet. */
#define HANDLE_LENGTH 2 /**< Length of handle inside FAS packet. */

/* TODO Consider changing the max values if encoded data for characteristic/descriptor is fixed length */ 
#define MAX_RANDOMNUMBERCLEAR_LEN (BLE_L2CAP_MTU_DEF - OPCODE_LENGTH - HANDLE_LENGTH) /**< Maximum size of a transmitted RandomNumberClear. */ 
#define MAX_RANDOMNUMBERCIPHER_LEN (BLE_L2CAP_MTU_DEF - OPCODE_LENGTH - HANDLE_LENGTH) /**< Maximum size of a transmitted RandomNumberCipher. */ 


/**@brief Function for encoding RandomNumberClear.
 *
 * @param[in]   p_randomnumberclear              RandomNumberClear characteristic structure to be encoded.
 * @param[out]  p_encoded_buffer   Buffer where the encoded data will be written.
 *
 * @return      Size of encoded data.
 */
static uint8_t randomnumberclear_encode(ble_fas_randomnumberclear_t * p_randomnumberclear, uint8_t * encoded_buffer)
{
    uint8_t len = 0;
    len += bds_uint16_encode(&p_randomnumberclear->random_num_generated, &encoded_buffer[len]);
    return len;
}

/**@brief Function for encoding RandomNumberCipher.
 *
 * @param[in]   p_randomnumbercipher              RandomNumberCipher characteristic structure to be encoded.
 * @param[out]  p_encoded_buffer   Buffer where the encoded data will be written.
 *
 * @return      Size of encoded data.
 */
static uint8_t randomnumbercipher_encode(ble_fas_randomnumbercipher_t * p_randomnumbercipher, uint8_t * encoded_buffer)
{
    uint8_t len = 0;
    len += bds_uint16_encode(&p_randomnumbercipher->aes_ecb_encrypted_value, &encoded_buffer[len]);
    return len;
}

/**@brief Function for decoding RandomNumberCipher.
 *
 * @param[in]   data_len              Length of the field to be decoded.
 * @param[in]   p_data                Buffer where the encoded data is stored.
 * @param[out]  p_write_val           Decoded data.
 *
 * @return      Length of the decoded field.
 */
static uint8_t randomnumbercipher_decode(uint8_t data_len, uint8_t * p_data, ble_fas_randomnumbercipher_t * p_write_val)
{
    uint8_t pos = 0;
    pos += bds_uint16_decode((data_len-pos), &p_data[pos], &p_write_val->aes_ecb_encrypted_value);

    return pos;
}

/**@brief Function for decoding RandomNumberClear.
 *
 * @param[in]   data_len              Length of the field to be decoded.
 * @param[in]   p_data                Buffer where the encoded data is stored.
 * @param[out]  p_write_val           Decoded data.
 *
 * @return      Length of the decoded field.
 */
static uint8_t randomnumberclear_decode(uint8_t data_len, uint8_t * p_data, ble_fas_randomnumberclear_t * p_write_val)
{
    uint8_t pos = 0;
    pos += bds_uint16_decode((data_len-pos), &p_data[pos], &p_write_val->random_num_generated);

    return pos;
}

/**@brief Function for handling the Connect event.
 *
 * @param[in]   p_fas       FAS Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
uint16_t rand_num_clear;
uint16_t calculated_cipher;
static void on_connect(ble_fas_t * p_fas, ble_evt_t * p_ble_evt)
{
  uint16_t err_code = 0;
  p_fas->conn_handle = p_ble_evt->evt.gap_evt.conn_handle;

  ble_fas_randomnumberclear_t rnc;
  rnc.random_num_generated = rand_num_clear + 1;
  err_code = ble_fas_randomnumberclear_set(p_fas, &rnc);
  APP_ERROR_CHECK(err_code);

   NRF_LOG_INFO("PERIPHERAL: Random Number Sent[0x%04x], Calculated Cipher[0x%04x]", rand_num_clear, calculated_cipher);

}

/**@brief Function for handling the Disconnect event.
 *
 * @param[in]   p_fas       FAS Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_disconnect(ble_fas_t * p_fas, ble_evt_t * p_ble_evt)
{
    UNUSED_PARAMETER(p_ble_evt);
    p_fas->conn_handle = BLE_CONN_HANDLE_INVALID;
}

/**@brief Function for handling the Write event.
 *
 * @param[in]   p_fas       FAS Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_write(ble_fas_t * p_fas, ble_gatts_evt_write_t * p_ble_evt)
{
    
  NRF_LOG_INFO("Handle = 0x%04x",p_ble_evt->handle);
    if(p_ble_evt->handle == p_fas->randomnumbercipher_handles.value_handle)
    {
        if(p_fas->evt_handler != NULL)
        {
            ble_fas_evt_t evt;
            evt.evt_type = BLE_FAS_RANDOMNUMBERCIPHER_EVT_WRITE;
            randomnumbercipher_decode(p_ble_evt->len, p_ble_evt->data, &evt.params.randomnumbercipher);
            NRF_LOG_INFO("Received Cipher = 0x%04x",evt.params.randomnumbercipher.aes_ecb_encrypted_value);
            //evt.characteristic = CIPHER_CHAR;
            p_fas->evt_handler(p_fas, &evt);
        }
    }
    if(p_ble_evt->handle == p_fas->randomnumberclear_handles.value_handle)
    {
        if(p_fas->evt_handler != NULL)
        {
            ble_fas_evt_t evt;
            evt.evt_type = BLE_FAS_RANDOMNUMBERCLEAR_EVT_WRITE;
            randomnumberclear_decode(p_ble_evt->len, p_ble_evt->data, &evt.params.randomnumbercipher);
            NRF_LOG_INFO("Received RNC = 0x%04x",evt.params.randomnumberclear.random_num_generated);
            //evt.characteristic = CIPHER_CHAR;
            p_fas->evt_handler(p_fas, &evt);
        }
    }
}

/**@brief Authorize WRITE request event handler.
 *
 * @details Handles WRITE events from the BLE stack.
 *
 * @param[in]   p_sc_ctrlpt  SC Ctrlpt structure.
 * @param[in]   p_gatts_evt  GATTS Event received from the BLE stack.
 *
 */
static void on_rw_authorize_request(ble_fas_t * p_fas, ble_gatts_evt_t * p_gatts_evt)
{
    ble_gatts_evt_rw_authorize_request_t * p_auth_req = &p_gatts_evt->params.authorize_request;
    uint32_t err_code;
    ble_gatts_rw_authorize_reply_params_t auth_reply;
    NRF_LOG_INFO("read.handle=%d, type=%d",p_auth_req->request.read.handle, p_auth_req->type);
    memset(&auth_reply, 0, sizeof(ble_gatts_rw_authorize_reply_params_t));

    if (p_auth_req->type == BLE_GATTS_AUTHORIZE_TYPE_WRITE)
      {
        auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_WRITE;

        if (   (p_gatts_evt->params.authorize_request.request.write.op
            != BLE_GATTS_OP_PREP_WRITE_REQ)
            && (p_gatts_evt->params.authorize_request.request.write.op
                != BLE_GATTS_OP_EXEC_WRITE_REQ_NOW)
                && (p_gatts_evt->params.authorize_request.request.write.op
                    != BLE_GATTS_OP_EXEC_WRITE_REQ_CANCEL)
        )
          {
            if ((p_auth_req->request.write.handle == p_fas->randomnumbercipher_handles.value_handle) ||
                (p_auth_req->request.write.handle == p_fas->randomnumberclear_handles.value_handle))
              {
                auth_reply.params.write.gatt_status = BLE_GATT_STATUS_SUCCESS;
                auth_reply.params.write.update = 1;
                auth_reply.params.write.offset = p_auth_req->request.write.offset;
                auth_reply.params.write.len = p_auth_req->request.write.len;
                auth_reply.params.write.p_data = p_auth_req->request.write.data;
                err_code = sd_ble_gatts_rw_authorize_reply(p_gatts_evt->conn_handle, &auth_reply);
                APP_ERROR_CHECK(err_code);

                on_write(p_fas, &p_auth_req->request.write);
              }
          }
      }

    else
      {
        if (p_auth_req->type == BLE_GATTS_AUTHORIZE_TYPE_READ)
          {
            auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_READ;
            if ((p_auth_req->request.read.handle == p_fas->randomnumbercipher_handles.value_handle) ||
                (p_auth_req->request.read.handle == p_fas->randomnumberclear_handles.value_handle))
              {
                  auth_reply.params.read.gatt_status = BLE_GATT_STATUS_SUCCESS;
                  err_code = sd_ble_gatts_rw_authorize_reply(p_gatts_evt->conn_handle, &auth_reply);
                  APP_ERROR_CHECK(err_code);
              }
          }
      }
}

/**@brief Function for handling BLE events.
 *
 * @param[in]   p_fas       FAS Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
void ble_fas_on_ble_evt(ble_fas_t * p_fas, ble_evt_t * p_ble_evt)
{
    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
        	//NRF_LOG_INFO("ble_fas_on_ble_evt() : BLE_GAP_EVT_CONNECTED");
            on_connect(p_fas, p_ble_evt);
            break;
        case BLE_GAP_EVT_DISCONNECTED:
            on_disconnect(p_fas, p_ble_evt);
            break;
        case BLE_GATTS_EVT_WRITE:
        	NRF_LOG_INFO("ble_fas_on_ble_evt() : BLE_GATTS_EVT_WRITE");
            on_write(p_fas, &p_ble_evt->evt.gatts_evt.params.write);
            break;
         case BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST:
           NRF_LOG_INFO("ble_fas_on_ble_evt() : BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST");
            on_rw_authorize_request(p_fas, &p_ble_evt->evt.gatts_evt);
            break;
        default:
            //No implementation needed.
            break;
    }
}

/**@brief Function for initializing the FAS. */
uint32_t ble_fas_init(ble_fas_t * p_fas, const ble_fas_init_t * p_fas_init)
{
    uint32_t err_code;
    ble_uuid_t ble_uuid;

    // Initialize service structure
    p_fas->evt_handler = p_fas_init->evt_handler;
    p_fas->conn_handle = BLE_CONN_HANDLE_INVALID;
    
    // Add a custom base UUID.
    ble_uuid128_t bds_base_uuid = {{0x5B, 0xCD, 0xFD, 0x5B, 0x43, 0xB7, 0xBE, 0x90, 0xFC, 0x4F, 0xAD, 0x22, 0x00, 0x00, 0xAC, 0x0D}};
    uint8_t       uuid_type;
    err_code = sd_ble_uuid_vs_add(&bds_base_uuid, &uuid_type);
    NRF_LOG_INFO("%s -> err_code [%d]",__func__, err_code);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }
    ble_uuid.type = uuid_type;
    ble_uuid.uuid = BLE_FAS_BASE_UUID;
        
    // Add service
    err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &ble_uuid, &p_fas->service_handle);
    NRF_LOG_INFO("%s -> err_code [%d]",__func__, err_code);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    } 

    // Add RandomNumberClear characteristic
    ble_fas_randomnumberclear_t randomnumberclear_initial_value = p_fas_init->ble_fas_randomnumberclear_initial_value;

    uint8_t randomnumberclear_encoded_value[MAX_RANDOMNUMBERCLEAR_LEN];
    ble_add_char_params_t add_randomnumberclear_params;
    memset(&add_randomnumberclear_params, 0, sizeof(add_randomnumberclear_params));
    
    add_randomnumberclear_params.uuid                = BLE_FAS_UUID_RAND_NUM_CLR;
    add_randomnumberclear_params.uuid_type           = ble_uuid.type; 
    add_randomnumberclear_params.max_len             = MAX_RANDOMNUMBERCLEAR_LEN;
    add_randomnumberclear_params.init_len            = randomnumberclear_encode(&randomnumberclear_initial_value, randomnumberclear_encoded_value);
    add_randomnumberclear_params.p_init_value        = randomnumberclear_encoded_value;
    add_randomnumberclear_params.char_props.read     = 1; 
    add_randomnumberclear_params.read_access         = SEC_OPEN; 
    add_randomnumberclear_params.char_props.write    = 1;
    add_randomnumberclear_params.is_defered_write    = 1;
    add_randomnumberclear_params.is_defered_read   = 1;
    add_randomnumberclear_params.write_access        = SEC_OPEN;
    // 1 for variable length and 0 for fixed length.
    add_randomnumberclear_params.is_var_len          = 1;
///

    err_code = characteristic_add(p_fas->service_handle, &add_randomnumberclear_params, &(p_fas->randomnumberclear_handles));
    NRF_LOG_INFO("%s -> err_code [%d]",__func__, err_code);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    } 

    // Add RandomNumberCipher characteristic
    ble_fas_randomnumbercipher_t randomnumbercipher_initial_value = p_fas_init->ble_fas_randomnumbercipher_initial_value;

    uint8_t randomnumbercipher_encoded_value[MAX_RANDOMNUMBERCIPHER_LEN];
    ble_add_char_params_t add_randomnumbercipher_params;
    memset(&add_randomnumbercipher_params, 0, sizeof(add_randomnumbercipher_params));
    
    add_randomnumbercipher_params.uuid                = BLE_FAS_UUID_CIPHOR_WRITE;
    add_randomnumbercipher_params.uuid_type           = ble_uuid.type; 
    add_randomnumbercipher_params.max_len             = MAX_RANDOMNUMBERCIPHER_LEN;
    add_randomnumbercipher_params.init_len            = randomnumbercipher_encode(&randomnumbercipher_initial_value, randomnumbercipher_encoded_value);
    add_randomnumbercipher_params.p_init_value        = randomnumbercipher_encoded_value;
    add_randomnumbercipher_params.char_props.read     = 1;
    add_randomnumbercipher_params.read_access         = SEC_OPEN;
    add_randomnumbercipher_params.char_props.write    = 1;
    add_randomnumbercipher_params.is_defered_write    = 1;
    add_randomnumbercipher_params.is_defered_read   = 1;
    add_randomnumbercipher_params.write_access        = SEC_OPEN;
    // 1 for variable length and 0 for fixed length.
    add_randomnumbercipher_params.is_var_len          = 1;

    err_code = characteristic_add(p_fas->service_handle, &add_randomnumbercipher_params, &(p_fas->randomnumbercipher_handles));
    NRF_LOG_INFO("%s -> err_code [%d]",__func__, err_code);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    } 

    return NRF_SUCCESS;
}

/**@brief Function for setting the RandomNumberClear. */
uint32_t ble_fas_randomnumberclear_set(ble_fas_t * p_fas, ble_fas_randomnumberclear_t * p_randomnumberclear)
{
    ble_gatts_value_t gatts_value;
    uint8_t encoded_value[MAX_RANDOMNUMBERCLEAR_LEN];

    // Initialize value struct.
    memset(&gatts_value, 0, sizeof(gatts_value));

    gatts_value.len     = randomnumberclear_encode(p_randomnumberclear, encoded_value);
    gatts_value.offset  = 0;
    gatts_value.p_value = encoded_value;

    return sd_ble_gatts_value_set(p_fas->conn_handle, p_fas->randomnumberclear_handles.value_handle, &gatts_value);
}

/**@brief Function for setting the RandomNumberClear. */
uint32_t ble_fas_randomnumberciphor_set(ble_fas_t * p_fas, ble_fas_randomnumbercipher_t * p_randomnumbercipher)
{
    ble_gatts_value_t gatts_value;
    uint8_t encoded_value[MAX_RANDOMNUMBERCIPHER_LEN];

    // Initialize value struct.
    memset(&gatts_value, 0, sizeof(gatts_value));

    gatts_value.len     = randomnumbercipher_encode(p_randomnumbercipher, encoded_value);
    gatts_value.offset  = 0;
    gatts_value.p_value = encoded_value;

    return sd_ble_gatts_value_set(p_fas->conn_handle, p_fas->randomnumbercipher_handles.value_handle, &gatts_value);
}
